__title__ = 'paxb'
__description__ = 'Python Architecture for XML Binding'
__url__ = 'https://github.com/dapper91/paxb'

__version__ = '0.2.1'

__author__ = 'Dmitry Pershin'
__email__ = 'dapper91@mail.ru'

__license__ = 'Public Domain License'
